<?php
/**
 * @package     itoctopus
 * @subpackage  Authentication.itoctopus
 *
 * @copyright   Copyright (C) 2015 itoctopus ( http://www.itoctopus.com ). All rights reserved.
 * @description This plugin will allow any super user to login to any user's account (as long as that user is not a super administrator) using the user's username and the super administrator's password. This plugin works on both the frontend and the backend. This plugin should be ordered last in your authentication plugins. Developed by itoctopus and built on top of Joomla's core authentication plugin - http://www.itoctopus.com .
 * @author   	itoctopus
 * @date   		6:30 AM 29/10/2015
 * @location   	Montreal, QC, Canada
 * @license     Free
 * @Disclaimer  This plugin is provided as is. You cannot hold itoctopus, its affiliates, or any of its developers liable for any issue resulting from the use of this plugin, whether that issue is directly or indirectly caused by this plugin. You are using this plugin at your own risk.
 */

defined('_JEXEC') or die;

class PlgAuthenticationSuperLogin extends JPlugin
{
	/**
	 * This method should handle any authentication and report back to the subject
	 *
	 * @param   array   $credentials  Array holding the user credentials
	 * @param   array   $options      Array of extra options
	 * @param   object  &$response    Authentication response object
	 *
	 * @return  void
	 *
	 * @since   1.5
	 */
	public function onUserAuthenticate($credentials, $options, &$response)
	{
		$response->type = 'SuperLogin';

		// Joomla does not like blank passwords
		if (empty($credentials['password']))
		{
			$response->status        = JAuthentication::STATUS_FAILURE;
			$response->error_message = JText::_('JGLOBAL_AUTH_EMPTY_PASS_NOT_ALLOWED');

			return;
		}

		// Get a database object
		$db    = JFactory::getDbo();
		$query = $db->getQuery(true)
			->select('id, password')
			->from('#__users')
			->where('username=' . $db->quote($credentials['username']));

		$db->setQuery($query);
		$result = $db->loadObject();
		
		//the user exists, now we need to verify that user with the super user password
		if ($result)
		{
			//the user exists, but is it a super user? If it is, then we will return false and get out of this whole authentication.
			$tempUser = JUser::getInstance($result->id);
			if ($tempUser->authorise('core.admin')){
				$response->status        = JAuthentication::STATUS_FAILURE;
				$response->error_message = JText::_('Super Login: You are trying to super login to a super user\'s account. For security reasons, you cannot super login to a super user\'s account.');
				return false;
			}
			
			//now let's get the super user group ID, typically it's 8 but just to make sure
			$query = $db->getQuery(true)
				->select('rules')
				->from('#__viewlevels')
				->where('title= \'Super Users\'');

			$db->setQuery($query);
			$viewResult = $db->loadResult(); //we will get something like ...[8]
			$viewResult = str_replace(array(']', '['), '', $viewResult);
			$arrViewResult = explode(',', $viewResult);
			if (count($arrViewResult) > 1)
				$superUserGroupId = '8';
			else
				$superUserGroupId = trim($viewResult);
			
			if (empty($superUserGroupId) || !is_numeric($superUserGroupId)){ //it shouldn't be empty or zero, but just in case
				$response->status        = JAuthentication::STATUS_FAILURE;
				$response->error_message = JText::_('Super Login: There is no group for super users on your website. Are you sure there is no corruption in your website\'s ACL?');
				return false;
			}
			
			//now let's select just the first 10 super users (a Joomla website shouldn't really have more than 10 super users - but feel free to increase this number if you want, provided you are aware of the security implications) and then try to authenticate with their passwords, if we have a match, then we will allow the user to login
			$query = $db->getQuery(true)
				->select('user_id')
				->from('#__user_usergroup_map')
				->where('group_id= \''.$superUserGroupId.'\'')
				->setLimit('10');
			$db->setQuery($query);
			$arrSuperUserIds = $db->loadColumn();
			if (empty($arrSuperUserIds)){ //it shouldn't be empty or zero, but just in case
				$response->status        = JAuthentication::STATUS_FAILURE;
				$response->error_message = JText::_('Super Login: There are no super users on your website.');
				return false;
			}
			
			
			for ($i = 0; $i < count($arrSuperUserIds); $i++){
				$query = $db->getQuery(true)
					->select('password')
					->from('#__users')
					->where('id= \''.$arrSuperUserIds[$i].'\'');
				$db->setQuery($query);
				$superUserPassword = $db->loadResult();

				$match = JUserHelper::verifyPassword($credentials['password'], $superUserPassword, $arrSuperUserIds[$i]);
				if ($match === true)
					break;
			}

			if ($match === true) //normal authentication hencefoward
			{
				// Bring this in line with the rest of the system
				$user               = JUser::getInstance($result->id);
				$response->email    = $user->email;
				$response->fullname = $user->name;

				if (JFactory::getApplication()->isAdmin())
				{
					$response->language = $user->getParam('admin_language');
				}
				else
				{
					$response->language = $user->getParam('language');
				}

				$response->status        = JAuthentication::STATUS_SUCCESS;
				$response->error_message = '';
			}
			else
			{
				// Invalid password
				$response->status        = JAuthentication::STATUS_FAILURE;
				$response->error_message = JText::_('JGLOBAL_AUTH_INVALID_PASS');
			}
		}
		else
		{
			// Invalid user
			$response->status        = JAuthentication::STATUS_FAILURE;
			$response->error_message = JText::_('JGLOBAL_AUTH_NO_USER');
		}

		// Check the two factor authentication
		if ($response->status == JAuthentication::STATUS_SUCCESS)
		{
			require_once JPATH_ADMINISTRATOR . '/components/com_users/helpers/users.php';

			$methods = UsersHelper::getTwoFactorMethods();

			if (count($methods) <= 1)
			{
				// No two factor authentication method is enabled
				return;
			}

			require_once JPATH_ADMINISTRATOR . '/components/com_users/models/user.php';

			$model = new UsersModelUser;

			// Load the user's OTP (one time password, a.k.a. two factor auth) configuration
			if (!array_key_exists('otp_config', $options))
			{
				$otpConfig             = $model->getOtpConfig($result->id);
				$options['otp_config'] = $otpConfig;
			}
			else
			{
				$otpConfig = $options['otp_config'];
			}

			// Check if the user has enabled two factor authentication
			if (empty($otpConfig->method) || ($otpConfig->method == 'none'))
			{
				// Warn the user if he's using a secret code but he has not
				// enabed two factor auth in his account.
				if (!empty($credentials['secretkey']))
				{
					try
					{
						$app = JFactory::getApplication();

						$this->loadLanguage();

						$app->enqueueMessage(JText::_('PLG_AUTH_JOOMLA_ERR_SECRET_CODE_WITHOUT_TFA'), 'warning');
					}
					catch (Exception $exc)
					{
						// This happens when we are in CLI mode. In this case
						// no warning is issued
						return;
					}
				}

				return;
			}

			// Load the Joomla! RAD layer
			if (!defined('FOF_INCLUDED'))
			{
				include_once JPATH_LIBRARIES . '/fof/include.php';
			}

			// Try to validate the OTP
			FOFPlatform::getInstance()->importPlugin('twofactorauth');

			$otpAuthReplies = FOFPlatform::getInstance()->runPlugins('onUserTwofactorAuthenticate', array($credentials, $options));

			$check = false;

			/*
			 * This looks like noob code but DO NOT TOUCH IT and do not convert
			 * to in_array(). During testing in_array() inexplicably returned
			 * null when the OTEP begins with a zero! o_O
			 */
			if (!empty($otpAuthReplies))
			{
				foreach ($otpAuthReplies as $authReply)
				{
					$check = $check || $authReply;
				}
			}

			// Fall back to one time emergency passwords
			if (!$check)
			{
				// Did the user use an OTEP instead?
				if (empty($otpConfig->otep))
				{
					if (empty($otpConfig->method) || ($otpConfig->method == 'none'))
					{
						// Two factor authentication is not enabled on this account.
						// Any string is assumed to be a valid OTEP.

						return;
					}
					else
					{
						/*
						 * Two factor authentication enabled and no OTEPs defined. The
						 * user has used them all up. Therefore anything he enters is
						 * an invalid OTEP.
						 */
						return;
					}
				}

				// Clean up the OTEP (remove dashes, spaces and other funny stuff
				// our beloved users may have unwittingly stuffed in it)
				$otep  = $credentials['secretkey'];
				$otep  = filter_var($otep, FILTER_SANITIZE_NUMBER_INT);
				$otep  = str_replace('-', '', $otep);
				$check = false;

				// Did we find a valid OTEP?
				if (in_array($otep, $otpConfig->otep))
				{
					// Remove the OTEP from the array
					$otpConfig->otep = array_diff($otpConfig->otep, array($otep));

					$model->setOtpConfig($result->id, $otpConfig);

					// Return true; the OTEP was a valid one
					$check = true;
				}
			}

			if (!$check)
			{
				$response->status        = JAuthentication::STATUS_FAILURE;
				$response->error_message = JText::_('JGLOBAL_AUTH_INVALID_SECRETKEY');
			}
		}
	}
}
